/*

    Copyright (c) 2018 - 2020 Analog Flavor EURL

    Contact us for support:
        http://www.analogflavor.com

-----------------------------------------------------------------------------
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    * Neither the name of Analog Flavor EURL nor 
      the names of its contributors may be used to endorse or promote products derived from
      this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
-----------------------------------------------------------------------------

*/

#ifndef wiener_mpod_lv_hv_snmp_h
#define wiener_mpod_lv_hv_snmp_h

#ifndef AF_BSP_API
    #if defined _WIN32 || defined __CYGWIN__
        #ifdef AF_BSP_DLL_EXPORTS /* defined if we are building the DLL (instead of using it) */
            #define AF_BSP_API __declspec(dllexport)
        #else
            #define AF_BSP_API __declspec(dllimport)
        #endif
    #else
        #ifdef AF_BSP_DLL_EXPORTS /* defined if we are building the DLL (instead of using it) */
            #if __GNUC__ >= 4
                #define AF_BSP_API __attribute__ ((visibility("default")))
            #else
                #define AF_BSP_API
            #endif
        #else
            #define AF_BSP_API
        #endif
    #endif
#endif

#include "wiener_mpod_lv_hv_types.h"

#ifdef __cplusplus
extern "C" {
#endif

    /* Creates the internal structure used by the library. The pointer to the structure is returned. It has to be passed to all other functions as "obj_ptr". */
    AF_BSP_API void *mpod_snmp_create();
    /* Deletes and invalidates the internal structure used by the library. */
    AF_BSP_API void mpod_snmp_delete(void *obj_ptr);
    /* Establish the connection to a Wiener MPOD crate. The IP address, the port index and the SNMP community have to be given.
        The usual values for the port_index is 161. The community_name should be "guru".
        If thread_polling_interval_ms > 0, the communication with the MPOD crate will be handled by a thread.

        The consequences are the following:
        - the response time of the access functions will be a lot shorter.
        - communication breakdowns will not slow down the calling function.
        - the set...() functions return immediately without error code.
        - the error string might not correspond to the last error.
        - this makes debugging a lot more difficult.

        Therefore, for debugging, the value should be set to 0.
        In normal operation the value for thread_polling_interval_ms should be >= 500.
    */
    AF_BSP_API int mpod_snmp_crateConnect(void *obj_ptr, const char *ip_address, int port_index, const char *community_name, int thread_polling_interval_ms_fast, int thread_polling_interval_ms_intermediate, int thread_polling_interval_ms_slow);
    /* Disconnects from the currently controlled MPOD crate. */
    AF_BSP_API void mpod_snmp_crateDisconnect(void *obj_ptr);
    /* Returns the connection status. */
    AF_BSP_API mpod_snmp_connection_status_type mpod_snmp_crateGetConnectionStatus(void *obj_ptr);
    /* Returns the connection status as string. */
    AF_BSP_API const char *mpod_snmp_crateGetConnectionStatusAsString(void *obj_ptr);
    /* Returns 1 if the connection has been established and 0 is an error occurred. */
    AF_BSP_API int mpod_snmp_crateCheckConnectionIsOk(void *obj_ptr);
    /* The returned string gives a hint on what went wrong in the previous function call. */
    AF_BSP_API const char *mpod_snmp_crateGetErrorString(void *obj_ptr);

    /* Returns 1 if the Wiener MIB file was loaded. */
    AF_BSP_API int mpod_snmp_checkIsMibFileLoaded(void *obj_ptr);
    /* Sets the timeout interval for the net-snmp library in milliseconds. The default value is 500 ms. */
    AF_BSP_API void mpod_snmp_snmpSetTimeoutInterval(void *obj_ptr, int interval_in_ms);

    /* Returns the number of slots that are occupied by LV or HV modules. */
    AF_BSP_API int mpod_snmp_crateGetNumberOfOccupiedSlots(void *obj_ptr);
    /* Returns the indices of the slots that are occupied by LV or HV modules. The last entry in the array is a -1. */
    AF_BSP_API const int *mpod_snmp_crateGetOccupiedSlots(void *obj_ptr);
    /* Returns the number of channels that have been detected for the module in slot "slot_index". */
    AF_BSP_API int mpod_snmp_moduleGetNumberOfChannels(void *obj_ptr, unsigned slot_index);
    /* This function retuns 1 if the module in "slot_index" is a HV module. */
    AF_BSP_API int mpod_snmp_moduleGetIsHvModule(void *obj_ptr, unsigned slot_index);

    /* The following functions are equivalent to SNMP commands. They are documented in mpod-snmp-guide.pdf. */
    AF_BSP_API double mpod_snmp_channelGetConfigMaxCurrent(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetConfigMaxSenseVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetConfigMaxTerminalVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetCurrent(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetCurrentFallRate(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetCurrentRiseRate(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API int mpod_snmp_channelGetIndex(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetMeasurementCurrent(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetMeasurementSenseVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetMeasurementTemperature(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetMeasurementTerminalVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API int mpod_snmp_channelGetStatus(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API const char *mpod_snmp_channelGetStatusAsString(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API int mpod_snmp_channelGetSupervisionBehavior(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetSupervisionMaxCurrent(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetSupervisionMaxPower(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetSupervisionMaxTerminalVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetSupervisionMinSenseVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API int mpod_snmp_channelGetSwitch(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API int mpod_snmp_channelGetTripTime(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API int mpod_snmp_channelGetUserConfig(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetVoltageFallRate(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API double mpod_snmp_channelGetVoltageRiseRate(void *obj_ptr, unsigned slot_index, unsigned channel_index);
    AF_BSP_API int mpod_snmp_channelSetCurrent(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_channelSetCurrentFallRate(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_channelSetCurrentRiseRate(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_channelSetSupervisionBehavior(void *obj_ptr, unsigned slot_index, unsigned channel_index, int value);
    AF_BSP_API int mpod_snmp_channelSetSupervisionMaxCurrent(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_channelSetSupervisionMaxTerminalVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_channelSetSupervisionMinSenseVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_channelSetSwitch(void *obj_ptr, unsigned slot_index, unsigned channel_index, int value);
    AF_BSP_API int mpod_snmp_channelSetTripTime(void *obj_ptr, unsigned slot_index, unsigned channel_index, int value);
    AF_BSP_API int mpod_snmp_channelSetUserConfig(void *obj_ptr, unsigned slot_index, unsigned channel_index, int value);
    AF_BSP_API int mpod_snmp_channelSetVoltage(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_channelSetVoltageFallRate(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_channelSetVoltageRiseRate(void *obj_ptr, unsigned slot_index, unsigned channel_index, double value);
    AF_BSP_API int mpod_snmp_crateGetFanSpeed(void *obj_ptr);
    AF_BSP_API int mpod_snmp_crateGetNumberOfChannels(void *obj_ptr);
    AF_BSP_API int mpod_snmp_crateGetNumberOfGroups(void *obj_ptr);
    AF_BSP_API int mpod_snmp_crateGetNumberOfModules(void *obj_ptr);
    AF_BSP_API const char *mpod_snmp_crateGetSerialNumber(void *obj_ptr);
    AF_BSP_API int mpod_snmp_crateGetSysMainSwitch(void *obj_ptr);
    AF_BSP_API const char *mpod_snmp_crateGetSysDescr(void *obj_ptr);
    AF_BSP_API int mpod_snmp_crateGetSysStatus(void *obj_ptr);
    AF_BSP_API const char *mpod_snmp_crateGetSysStatusAsString(void *obj_ptr);
    AF_BSP_API int mpod_snmp_crateSetFanSpeed(void *obj_ptr, int value);
    AF_BSP_API int mpod_snmp_crateSetSysHardwareReset(void *obj_ptr, int value);
    AF_BSP_API int mpod_snmp_crateSetSysMainSwitch(void *obj_ptr, int value);
    AF_BSP_API int mpod_snmp_groupGetIndex(void *obj_ptr, unsigned index);
    AF_BSP_API int mpod_snmp_groupSetSwitch(void *obj_ptr, unsigned group_index, int value);
    AF_BSP_API double mpod_snmp_moduleGetAuxiliaryMeasurementVoltage0(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetAuxiliaryMeasurementVoltage1(void *obj_ptr, unsigned slot_index);
    AF_BSP_API const char *mpod_snmp_moduleGetDescription(void *obj_ptr, unsigned slot_index);
    AF_BSP_API int mpod_snmp_moduleGetEventChannelStatus(void *obj_ptr, unsigned slot_index);
    AF_BSP_API int mpod_snmp_moduleGetEventStatus(void *obj_ptr, unsigned slot_index);
    AF_BSP_API const char *mpod_snmp_moduleGetEventStatusAsString(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetHardwareLimitCurrent(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetHardwareLimitVoltage(void *obj_ptr, unsigned slot_index);
    AF_BSP_API int mpod_snmp_moduleGetIndex(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetMeasurementTemperature0(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetMeasurementTemperature1(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetMeasurementTemperature2(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetMeasurementTemperature3(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetRampSpeedCurrent(void *obj_ptr, unsigned slot_index);
    AF_BSP_API double mpod_snmp_moduleGetRampSpeedVoltage(void *obj_ptr, unsigned slot_index);
    AF_BSP_API int mpod_snmp_moduleGetStatus(void *obj_ptr, unsigned slot_index);
    AF_BSP_API const char *mpod_snmp_moduleGetStatusAsString(void *obj_ptr, unsigned slot_index);
    AF_BSP_API int mpod_snmp_moduleSetDoClear(void *obj_ptr, unsigned slot_index, int value);
    AF_BSP_API int mpod_snmp_moduleSetRampSpeedCurrent(void *obj_ptr, unsigned slot_index, double value);
    AF_BSP_API int mpod_snmp_moduleSetRampSpeedVoltage(void *obj_ptr, unsigned slot_index, double value);

    /* The following functions allow to directly access the SNMP protocol. */
    AF_BSP_API const char *mpod_snmp_snmpGetDescription(void *obj_ptr, const char *string_oid);
    AF_BSP_API const char *mpod_snmp_snmpGetNextOid(void *obj_ptr, const char *string_oid);
    AF_BSP_API double mpod_snmp_snmpGetValueDouble(void *obj_ptr, const char *string_oid);
    AF_BSP_API int mpod_snmp_snmpGetValueInt(void *obj_ptr, const char *string_oid);
    AF_BSP_API const char *mpod_snmp_snmpGetValueString(void *obj_ptr, const char *string_oid);
    AF_BSP_API int mpod_snmp_snmpSetValueDouble(void *obj_ptr, const char *string_oid, double value);
    AF_BSP_API int mpod_snmp_snmpSetValueInt(void *obj_ptr, const char *string_oid, int value);
    AF_BSP_API const char *mpod_snmp_snmpTranslateFromOid(void *obj_ptr, const char *string_oid);
    AF_BSP_API const char *mpod_snmp_snmpTranslateToOid(void *obj_ptr, const char *string_oid);

    AF_BSP_API const char *mpod_snmp_convertCrateStatusToString(void *obj_ptr, int integer_status);
    AF_BSP_API const char *mpod_snmp_convertModuleEventStatusToString(void *obj_ptr, int integer_status);
    AF_BSP_API const char *mpod_snmp_convertModuleStatusToString(void *obj_ptr, int integer_status);
    AF_BSP_API const char *mpod_snmp_convertChannelStatusToString(void *obj_ptr, int integer_status, unsigned slot_index);

    /* Windows only, has no effect on Linux */
    /* this is not needed if the integrating application calls the Windows functions WSAStartup() and WSACleanup() */
    AF_BSP_API int mpod_snmp_winsock_startup();
    AF_BSP_API void mpod_snmp_winsock_cleanup();

#ifdef __cplusplus
}
#endif


/*    wiener_mpod_lv_hv_snmp_h */
#endif   


